"""
@file
@brief Documentation helper.
"""
import keyword
import textwrap
import re
from jinja2 import Template
from jinja2.runtime import Undefined
from onnx.defs import OpSchema
from ...tools import change_style


def type_mapping(name):
    """
    Mapping between types name and type integer value.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.doc.doc_helper import type_mapping
        import pprint
        pprint.pprint(type_mapping(None))
        print(type_mapping("INT"))
        print(type_mapping(2))
    """
    di = dict(FLOAT=1, FLOATS=6, GRAPH=5, GRAPHS=10, INT=2,
              INTS=7, STRING=3, STRINGS=8, TENSOR=4,
              TENSORS=9, UNDEFINED=0, SPARSE_TENSOR=11)
    if name is None:
        return di
    if isinstance(name, str):
        return di[name]
    rev = {v: k for k, v in di.items()}
    return rev[name]


def _get_doc_template():

    return Template(textwrap.dedent("""
        {% for sch in schemas %}

        {{format_name_with_domain(sch)}}
        {{'=' * len(format_name_with_domain(sch))}}

        {{process_documentation(sch.doc)}}

        {% if sch.attributes %}
        **Attributes**

        {% for _, attr in sorted(sch.attributes.items()) %}* *{{attr.name}}*{%
          if attr.required %} (required){% endif %}: {{
          process_attribute_doc(attr.description)}} {%
          if attr.default_value %} {{
          process_default_value(attr.default_value)
          }} ({{type_mapping(attr.type)}}){% endif %}
        {% endfor %}
        {% endif %}

        {% if sch.inputs %}
        **Inputs**

        {% if sch.min_input != sch.max_input %}Between {{sch.min_input
        }} and {{sch.max_input}} inputs.
        {% endif %}
        {% for ii, inp in enumerate(sch.inputs) %}
        * *{{getname(inp, ii)}}*{{format_option(inp)}}{{inp.typeStr}}: {{
        inp.description}}{% endfor %}
        {% endif %}

        {% if sch.outputs %}
        **Outputs**

        {% if sch.min_output != sch.max_output %}Between {{sch.min_output
        }} and {{sch.max_output}} outputs.
        {% endif %}
        {% for ii, out in enumerate(sch.outputs) %}
        * *{{getname(out, ii)}}*{{format_option(out)}}{{out.typeStr}}: {{
        out.description}}{% endfor %}
        {% endif %}

        {% if sch.type_constraints %}
        **Type Constraints**

        {% for ii, type_constraint in enumerate(sch.type_constraints)
        %}* {{getconstraint(type_constraint, ii)}}: {{
        type_constraint.description}}
        {% endfor %}
        {% endif %}

        **Version**

        *Onnx name:* `{{sch.name}} <{{build_doc_url(sch)}}{{sch.name}}>`_

        {% if sch.support_level == OpSchema.SupportType.EXPERIMENTAL %}
        No versioning maintained for experimental ops.
        {% else %}
        This version of the operator has been {% if
        sch.deprecated %}deprecated{% else %}available{% endif %} since
        version {{sch.since_version}}{% if
        sch.domain %} of domain {{sch.domain}}{% endif %}.
        {% if len(sch.versions) > 1 %}
        Other versions of this operator:
        {% for v in sch.version[:-1] %} {{v}} {% endfor %}
        {% endif %}
        {% endif %}

        **Runtime implementation:**
        :class:`{{sch.name}}
        <mlprodict.onnxrt.ops_cpu.op_{{change_style(sch.name)}}.{{sch.name}}>`

        {% endfor %}
    """))


_template_operator = _get_doc_template()


class NewOperatorSchema:
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleRegressorDouble.
    """

    def __init__(self, name):
        self.name = name
        self.domain = 'mlprodict'


def get_rst_doc(op_name):
    """
    Returns a documentation in RST format
    for all :class:`OnnxOperator`.

    :param op_name: operator name of None for all
    :return: string

    The function relies on module :epkg:`jinja2` or replaces it
    with a simple rendering if not present.
    """
    from ..ops_cpu._op import _schemas
    schemas = [_schemas.get(op_name, NewOperatorSchema(op_name))]

    def format_name_with_domain(sch):
        if sch.domain:
            return '{} ({})'.format(sch.name, sch.domain)
        return sch.name

    def format_option(obj):
        opts = []
        if OpSchema.FormalParameterOption.Optional == obj.option:
            opts.append('optional')
        elif OpSchema.FormalParameterOption.Variadic == obj.option:
            opts.append('variadic')
        if getattr(obj, 'isHomogeneous', False):
            opts.append('heterogeneous')
        if opts:
            return " (%s)" % ", ".join(opts)
        return ""  # pragma: no cover

    def getconstraint(const, ii):
        if const.type_param_str:
            name = const.type_param_str
        else:
            name = str(ii)  # pragma: no cover
        if const.allowed_type_strs:
            name += " " + ", ".join(const.allowed_type_strs)
        return name

    def getname(obj, i):
        name = obj.name
        if len(name) == 0:
            return str(i)  # pragma: no cover
        return name

    def process_documentation(doc):
        if doc is None:
            doc = ''  # pragma: no cover
        if isinstance(doc, Undefined):
            doc = ''  # pragma: no cover
        if not isinstance(doc, str):
            raise TypeError(  # pragma: no cover
                "Unexpected type {} for {}".format(type(doc), doc))
        doc = textwrap.dedent(doc)
        main_docs_url = "https://github.com/onnx/onnx/blob/master/"
        rep = {
            '[the doc](IR.md)': '`ONNX <{0}docs/IR.md>`_',
            '[the doc](Broadcasting.md)':
                '`Broadcasting in ONNX <{0}docs/Broadcasting.md>`_',
            '<dl>': '',
            '</dl>': '',
            '<dt>': '* ',
            '<dd>': '  ',
            '</dt>': '',
            '</dd>': '',
            '<tt>': '``',
            '</tt>': '``',
            '<br>': '\n',
        }
        for k, v in rep.items():
            doc = doc.replace(k, v.format(main_docs_url))
        move = 0
        lines = []
        for line in doc.split('\n'):
            if line.startswith("```"):
                if move > 0:
                    move -= 4
                    lines.append("\n")
                else:
                    lines.append("::\n")
                    move += 4
            elif move > 0:
                lines.append(" " * move + line)
            else:
                lines.append(line)
        return "\n".join(lines)

    def process_attribute_doc(doc):
        return doc.replace("<br>", " ")

    def build_doc_url(sch):
        doc_url = "https://github.com/onnx/onnx/blob/master/docs/Operators"
        if "ml" in sch.domain:
            doc_url += "-ml"
        doc_url += ".md"
        doc_url += "#"
        if sch.domain not in (None, '', 'ai.onnx'):
            doc_url += sch.domain + "."
        return doc_url

    def process_default_value(value):
        if value is None:
            return ''  # pragma: no cover
        res = []
        for c in str(value):
            if ((c >= 'A' and c <= 'Z') or (c >= 'a' and c <= 'z') or
                    (c >= '0' and c <= '9')):
                res.append(c)
                continue
            if c in '[]-+(),.?':
                res.append(c)
                continue
        if len(res) == 0:
            return "*default value cannot be automatically retrieved*"
        return "Default value is ``" + ''.join(res) + "``"

    fnwd = format_name_with_domain
    tmpl = _template_operator
    docs = tmpl.render(schemas=schemas, OpSchema=OpSchema,
                       len=len, getattr=getattr, sorted=sorted,
                       format_option=format_option,
                       getconstraint=getconstraint,
                       getname=getname, enumerate=enumerate,
                       format_name_with_domain=fnwd,
                       process_documentation=process_documentation,
                       build_doc_url=build_doc_url, str=str,
                       type_mapping=type_mapping,
                       process_attribute_doc=process_attribute_doc,
                       process_default_value=process_default_value,
                       change_style=change_style)
    return docs.replace(" Default value is ````", "")


def debug_onnx_object(obj, depth=3):
    """
    ``__dict__`` is not in most of :epkg:`onnx` objects.
    This function uses function *dir* to explore this object.
    """
    def iterable(o):
        try:
            iter(o)
            return True
        except TypeError:
            return False

    if depth <= 0:
        return None

    rows = [str(type(obj))]
    if not isinstance(obj, (int, str, float, bool)):

        for k in sorted(dir(obj)):
            try:
                val = getattr(obj, k)
                sval = str(val).replace("\n", " ")
            except (AttributeError, ValueError) as e:  # pragma: no cover
                sval = "ERRROR-" + str(e)
                val = None

            if 'method-wrapper' in sval or "built-in method" in sval:
                continue

            rows.append("- {}: {}".format(k, sval))
            if k.startswith('__') and k.endswith('__'):
                continue
            if val is None:
                continue

            if isinstance(val, dict):
                try:
                    sorted_list = list(sorted(val.items()))
                except TypeError:  # pragma: no cover
                    sorted_list = list(val.items())
                for kk, vv in sorted_list:
                    rows.append("  - [%s]: %s" % (str(kk), str(vv)))
                    res = debug_onnx_object(vv, depth - 1)
                    if res is None:
                        continue
                    for line in res.split("\n"):
                        rows.append("    " + line)
            elif iterable(val):
                if all(map(lambda o: isinstance(o, (str, bytes)) and len(o) == 1, val)):
                    continue
                for i, vv in enumerate(val):
                    rows.append("  - [%d]: %s" % (i, str(vv)))
                    res = debug_onnx_object(vv, depth - 1)
                    if res is None:
                        continue
                    for line in res.split("\n"):
                        rows.append("    " + line)
            elif not callable(val):
                res = debug_onnx_object(val, depth - 1)
                if res is None:
                    continue
                for line in res.split("\n"):
                    rows.append("  " + line)

    return "\n".join(rows)


def visual_rst_template():
    """
    Returns a :epkg:`jinja2` template to display DOT graph for each
    converter from :epkg:`sklearn-onnx`.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.doc.doc_helper import visual_rst_template
        print(visual_rst_template())
    """
    return textwrap.dedent("""

    .. _l-{{link}}:

    {{ title }}
    {{ '=' * len(title) }}

    Fitted on a problem type *{{ kind }}*
    (see :func:`find_suitable_problem
    <mlprodict.onnxrt.validate.validate_problems.find_suitable_problem>`),
    method {{ method }} matches output {{ output_index }}.
    {{ optim_param }}

    ::

        {{ indent(model, "    ") }}

    {{ table }}

    .. gdot::

        {{ indent(dot, "    ") }}
    """)



"""
@file
@brief One class which visits a syntax tree.
"""
import inspect
import ast
from textwrap import dedent
import numpy
from scipy.spatial.distance import squareform, pdist
from .node_visitor_translator import CodeNodeVisitor


def py_make_float_array(cst, op_version=None):
    """
    Creates an array with a single element
    from a constant.

    @param      cst         constant
    @param      op_version  unused
    @return                 array

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnx_grammar.onnx_translation import py_make_float_array
        print(py_make_float_array(5.5))
    """
    return numpy.array([cst], dtype=numpy.float32)


def py_pow(x, p, op_version=None):
    """
    Function for python operator ``**``.

    @param      x           float
    @param      p           power
    @param      op_version  unused
    @return                 :math:`x^p`
    """
    return x ** p


def py_mul(*x, op_version=None):
    """
    Function for python operator ``*``.

    @param      x           floats
    @param      op_version  unused
    @return                 `x*y`
    """
    if len(x) == 2:
        return x[0] * x[1]
    p = x[0]
    for y in x[1:]:
        p *= y
    return p


def py_opp(x, op_version=None):
    """
    Function for python unary operator ``-``.

    @param      x           floats
    @param      op_version  unused
    @return                 `-x`
    """
    return -x


def squareform_pdist(X, metric='sqeuclidean', op_version=None):
    """
    Replacements for `squareform
    <http://scipy.github.io/devdocs/generated/scipy.spatial.distance.squareform.html>`_
    and `pdist
    <http://scipy.github.io/devdocs/generated/scipy.spatial.distance.pdist.html>`_.
    """
    return squareform(pdist(X, metric=metric))


def get_default_context():
    """
    Returns a default context useful for most of the conversion
    from a function using :epkg:`numpy` into :epkg:`ONNX`.
    """
    context = {'py_pow': py_pow, 'py_make_float_array': py_make_float_array,
               'py_mul': py_mul, 'py_opp': py_opp,
               'cdist': 'cdist', 'squareform_pdist': 'squareform_pdist'}
    allow = set(('abs add ceil arccos arccosh arcsin arcsinh arctan arctanh ceil cos cosh divide'
                 'equal exp floor greater invert less log matmul maximum minimum mod'
                 'multiply power sign sin sinh sqrt square subtract tan tanh transpose').split())
    for k, v in numpy.__dict__.items():
        if k not in allow:
            continue
        context['numpy.%s' % k] = v
        context['np.%s' % k] = v
    return context


def get_default_context_cpl():
    """
    Returns a default useful context to compile the converter
    returned by @see fn translate_fct2onnx.
    """
    ctx = {'py_make_float_array': py_make_float_array,
           'py_pow': py_pow, 'py_mul': py_mul, 'py_opp': py_opp,
           'numpy': numpy}
    try:
        from skl2onnx.algebra.complex_functions import onnx_squareform_pdist
        from skl2onnx.algebra.complex_functions import onnx_cdist
        ctx['onnx_squareform_pdist'] = onnx_squareform_pdist
        ctx['onnx_cdist'] = onnx_cdist
    except ImportError:  # pragma: no cover
        # Too old version for skl2onnx.
        pass

    from skl2onnx.algebra import onnx_ops
    from skl2onnx.algebra.onnx_operator import OnnxOperator
    d = onnx_ops.__dict__
    for k, v in d.items():
        try:
            if k.startswith("Onnx") and issubclass(v, OnnxOperator):
                ctx[k] = v
        except TypeError as e:
            if inspect.isfunction(v):
                continue
            raise RuntimeError(  # pragma: no cover
                "Issue with {}={} (type={})".format(k, v, type(v))) from e
    return ctx


def translate_fct2onnx(fct, context=None, cpl=False,
                       context_cpl=None, output_names=None,
                       dtype=numpy.float32,
                       verbose=0, fLOG=None):
    """
    Translates a function into :epkg:`ONNX`. The code it produces
    is using classes *OnnxAbs*, *OnnxAdd*, ...

    @param      fct             function to convert
    @param      context         context of the function to convert
                                something like ``{'numpy.transpose': numpy.transpose}``,
                                if *context* is None, it receives a default value
                                returnd by @see fn get_default_context
    @param      cpl             compile the function after it was
                                created
    @param      context_cpl     context used at compiling time
                                if *context_cpl* is None, it receives a default value
                                returnd by @see fn get_default_context_cpl
    @param      output_names    names of the output in the :epkg:`ONNX` graph
    @param      dtype           :epkg:`numpy` float type used to produce the model
    @param      verbose         integer, display more information
    @param      fLOG            logging function
    @return                     code or compiled code

    .. exref::
        :title: Convert a function into ONNX code

        The following code parses a python function and returns
        another python function which produces an :epkg:`ONNX`
        graph if executed.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning
            :process:
            :store_in_file: fct2onnx2.py

            import numpy
            from mlprodict.onnx_grammar import translate_fct2onnx

            def trs(x, y):
                z = x + numpy.transpose(y, axes=[1, 0])
                return x * z

            onnx_code = translate_fct2onnx(
                trs, context={'numpy.transpose': numpy.transpose})
            print(onnx_code)

    Next example goes further and compile the outcome.

    .. exref::
        :title: Convert a function into ONNX code and run

        The following code parses a python function and returns
        another python function which produces an :epkg:`ONNX`
        graph if executed. The example executes the function,
        creates an :epkg:`ONNX` then uses @see cl OnnxInference
        to compute *predictions*. Finally it compares
        them to the original.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning
            :process:
            :store_in_file: fct2onnx3.py

            import numpy
            from mlprodict.onnx_grammar import translate_fct2onnx
            from mlprodict.onnxrt import OnnxInference
            from skl2onnx.algebra.onnx_ops import (
                OnnxAdd, OnnxTranspose, OnnxMul, OnnxIdentity
            )

            ctx = {'OnnxAdd': OnnxAdd,
                   'OnnxTranspose': OnnxTranspose,
                   'OnnxMul': OnnxMul,
                   'OnnxIdentity': OnnxIdentity}

            def trs(x, y):
                z = x + numpy.transpose(y, axes=[1, 0])
                return x * z

            inputs = {'x': numpy.array([[1, 2]], dtype=numpy.float32),
                      'y': numpy.array([[-0.3, 0.4]], dtype=numpy.float32).T}

            original = trs(inputs['x'], inputs['y'])

            print('original output:', original)

            onnx_fct = translate_fct2onnx(
                trs, context={'numpy.transpose': numpy.transpose},
                cpl=True, context_cpl=ctx, output_names=['Z'])

            onnx_code = onnx_fct('x', 'y', opset_version=12)
            print('ONNX code:', onnx_code)

            onnx_g = onnx_code.to_onnx(inputs, target_opset=12)

            oinf = OnnxInference(onnx_g)
            res = oinf.run(inputs)

            print("ONNX inference:", res['Z'])
            print("ONNX graph:", onnx_g)

    The function to be converted may include python functions
    which must not be converted. In that case, their name
    must be prefixed by ``py_``. The execution of the function
    this one builds produces the following error::

        TypeError: Parameter to MergeFrom() must be instance of same class:
        expected onnx.TensorProto got onnx.AttributeProto.

    It indicates that constants in the code marges multiple types,
    usually floats and tensor of floats. Floats should be converted
    using the following function::

        def py_make_float_array(cst):
            return numpy.array([cst], dtype=numpy.float32)

    The function replaces empty contexts by default values which
    covers many :epkg:`numpy` functions. The tutorial
    :ref:`l-onnx-tutorial` gives an example of how it can be used
    on a more complex function.
    """
    def compile_code(name, code, context=None):
        """
        Compiles a python function with the given
        context.

        @param      name        function name
        @param      code        python code
        @param      context     context used at compilation
        @return                 compiled function
        """
        if context is None:
            context = {}  # pragma: no cover
        try:
            obj = compile(code, "", "exec")
        except SyntaxError as e:  # pragma: no cover
            raise SyntaxError("Unable to compile\n{}".format(code)) from e
        context_g = context.copy()
        context_l = context.copy()
        exec(obj, context_g, context_l)  # pylint: disable=W0122
        return context_l[name]

    if isinstance(fct, str):
        code = fct
    elif callable(fct):
        code = inspect.getsource(fct)
    else:
        raise TypeError(  # pragma: no cover
            "Unable to guess code from type {}.".format(type(fct)))
    node = ast.parse(dedent(code))
    v = CodeNodeVisitor()
    v.visit(node)
    if context is None:
        context = get_default_context()
    onnx_code = v.export(context=context,
                         output_names=output_names)
    if not cpl:
        return onnx_code
    if verbose > 0 and fLOG is not None:  # pragma: no cover
        fLOG('[translate_fct2onnx] python code')
        fLOG(code)
        fLOG('[translate_fct2onnx] ONNX code')
        fLOG(onnx_code)
    if context_cpl is None:
        context_cpl = get_default_context_cpl()
    if 'numpy' not in context_cpl:
        context_cpl = context_cpl.copy()
        context_cpl['numpy'] = numpy
    return compile_code(fct.__name__, onnx_code, context_cpl)



"""
@file
@brief A pipeline which serializes into ONNX steps by steps.
"""
import numpy
from sklearn.base import clone
from sklearn.pipeline import Pipeline, _fit_transform_one
from sklearn.utils.validation import check_memory
from sklearn.utils import _print_elapsed_time
from ..onnx_conv import to_onnx
from .onnx_transformer import OnnxTransformer


class OnnxPipeline(Pipeline):
    """
    The pipeline overwrites method *fit*, it trains and converts
    every steps into ONNX before training the next step
    in order to minimize discrepencies. By default,
    ONNX is using float and not double which is the default
    for :epkg:`scikit-learn`. It may introduce discrepencies
    when a non-continuous model (mathematical definition) such
    as tree ensemble and part of the pipeline.

    :param steps:
        List of (name, transform) tuples (implementing fit/transform) that are
        chained, in the order in which they are chained, with the last object
        an estimator.
    :param memory: str or object with the joblib.Memory interface, default=None
        Used to cache the fitted transformers of the pipeline. By default,
        no caching is performed. If a string is given, it is the path to
        the caching directory. Enabling caching triggers a clone of
        the transformers before fitting. Therefore, the transformer
        instance given to the pipeline cannot be inspected
        directly. Use the attribute ``named_steps`` or ``steps`` to
        inspect estimators within the pipeline. Caching the
        transformers is advantageous when fitting is time consuming.
    :param verbose: bool, default=False
        If True, the time elapsed while fitting each step will be printed as it
        is completed.
    :param output_name: string
        requested output name or None to request all and
        have method *transform* to store all of them in a dataframe
    :param enforce_float32: boolean
        :epkg:`onnxruntime` only supports *float32*,
        :epkg:`scikit-learn` usually uses double floats, this parameter
        ensures that every array of double floats is converted into
        single floats
    :param runtime: string, defined the runtime to use
        as described in @see cl OnnxInference.
    :param options: see @see fn to_onnx
    :param white_op: see @see fn to_onnx
    :param black_op: see @see fn to_onnx
    :param final_types: see @see fn to_onnx
    :param op_version: ONNX targeted opset

    The class stores transformers before converting them into ONNX
    in attributes ``raw_steps_``.

    See notebook :ref:`onnxdiscrepenciesrst` to see it can
    be used to reduce discrepencies after it was converted into
    *ONNX*.
    """

    def __init__(self, steps, *, memory=None, verbose=False,
                 output_name=None, enforce_float32=True,
                 runtime='python', options=None,
                 white_op=None, black_op=None, final_types=None,
                 op_version=None):
        self.output_name = output_name
        self.enforce_float32 = enforce_float32
        self.runtime = runtime
        self.options = options
        self.white_op = white_op
        self.white_op = white_op
        self.black_op = black_op
        self.final_types = final_types
        self.op_version = op_version
        # The constructor calls _validate_step and it checks the value
        # of black_op.
        Pipeline.__init__(
            self, steps, memory=memory, verbose=verbose)

    def fit(self, X, y=None, **fit_params):
        """
        Fits the model, fits all the transforms one after the
        other and transform the data, then fit the transformed
        data using the final estimator.

        :param X: iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.
        :param y: iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.
        :param fit_params: dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.
        :return: self, Pipeline, this estimator
        """
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt = self._fit(X, y, **fit_params_steps)
        with _print_elapsed_time('OnnxPipeline',
                                 self._log_message(len(self.steps) - 1)):
            if self._final_estimator != 'passthrough':
                fit_params_last_step = fit_params_steps[self.steps[-1][0]]
                self._final_estimator.fit(Xt, y, **fit_params_last_step)

        return self

    def _fit(self, X, y=None, **fit_params_steps):
        # shallow copy of steps - this should really be steps_
        if hasattr(self, 'raw_steps_') and self.raw_steps_ is not None:  # pylint: disable=E0203
            # Let's reuse the previous training.
            self.steps = list(self.raw_steps_)  # pylint: disable=E0203
            self.raw_steps_ = list(self.raw_steps_)
        else:
            self.steps = list(self.steps)
            self.raw_steps_ = list(self.steps)

        self._validate_steps()
        # Setup the memory
        memory = check_memory(self.memory)

        fit_transform_one_cached = memory.cache(_fit_transform_one)

        for (step_idx,
             name,
             transformer) in self._iter(with_final=False,
                                        filter_passthrough=False):
            if (transformer is None or transformer == 'passthrough'):
                with _print_elapsed_time('Pipeline',
                                         self._log_message(step_idx)):
                    continue

            if hasattr(memory, 'location'):
                # joblib >= 0.12
                if memory.location is None:
                    # we do not clone when caching is disabled to
                    # preserve backward compatibility
                    cloned_transformer = transformer
                else:
                    cloned_transformer = clone(transformer)
            else:
                cloned_transformer = clone(transformer)

            # Fit or load from cache the current transformer
            x_train = X
            X, fitted_transformer = fit_transform_one_cached(
                cloned_transformer, X, y, None,
                message_clsname='Pipeline',
                message=self._log_message(step_idx),
                **fit_params_steps[name])
            # Replace the transformer of the step with the fitted
            # transformer. This is necessary when loading the transformer
            # from the cache.
            self.raw_steps_[step_idx] = (name, fitted_transformer)
            self.steps[step_idx] = (
                name, self._to_onnx(name, fitted_transformer, x_train))
        return X

    def _to_onnx(self, name, fitted_transformer, x_train):
        """
        Converts a transformer into ONNX.

        @param  name                model name
        @param  fitted_transformer  fitted transformer
        @param  x_train             training dataset
        @return                     corresponding @see cl OnnxTransformer
        """
        if not isinstance(x_train, numpy.ndarray):
            raise RuntimeError(  # pragma: no cover
                "The pipeline only handle numpy arrays not {}.".format(
                    type(x_train)))
        atts = {'options', 'white_op', 'black_op', 'final_types'}
        kwargs = {k: getattr(self, k) for k in atts}
        if self.enforce_float32 or x_train.dtype != numpy.float64:
            x_train = x_train.astype(numpy.float32)
        if 'options' in kwargs:
            kwargs['options'] = self._preprocess_options(
                name, kwargs['options'])
        kwargs['target_opset'] = self.op_version
        onx = to_onnx(fitted_transformer, x_train, **kwargs)
        tr = OnnxTransformer(
            onx.SerializeToString(), output_name=self.output_name,
            enforce_float32=self.enforce_float32, runtime=self.runtime)
        return tr.fit()

    def _preprocess_options(self, name, options):
        """
        Preprocesses the options.

        @param      name        option name
        @param      options     conversion options
        @return                 new options
        """
        if options is None:
            return None
        prefix = name + '__'
        new_options = {}
        for k, v in options.items():
            if isinstance(k, str):
                if k.startswith(prefix):
                    new_options[k[len(prefix):]] = v
            else:
                new_options[k] = v
        return new_options



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.defs import onnx_opset_version
from ..shape_object import ShapeObject
from ._op import OpRunUnaryNum, OpRun


class Unsqueeze_1(OpRunUnaryNum):

    atts = {'axes': [], 'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Unsqueeze_1.atts,
                               **options)
        if isinstance(self.axes, numpy.ndarray):
            self.axes = tuple(self.axes)
        elif self.axes in [[], tuple()]:
            self.axes = None
        elif isinstance(self.axes, list):
            self.axes = tuple(self.axes)

    def _run(self, data):  # pylint: disable=W0221
        if isinstance(self.axes, (tuple, list)):
            sq = data
            for a in self.axes:
                sq = numpy.expand_dims(sq, axis=a)
        else:
            raise RuntimeError(  # pragma: no cover
                "axes cannot be None for operator Unsqueeze (Unsqueeze_1).")
        return (sq, )

    def _infer_shapes(self, x):  # pylint: disable=W0221
        return (x.unsqueeze(axes=self.axes), )


class Unsqueeze_11(Unsqueeze_1):
    pass


class Unsqueeze_13(OpRun):

    atts = {'keepdims': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Unsqueeze_13.atts,
                       **options)
        self.axes = None

    def _run(self, data, axes=None):  # pylint: disable=W0221
        if axes is not None:
            sq = data
            for a in axes:
                sq = numpy.expand_dims(sq, axis=a)
        else:
            raise RuntimeError(  # pragma: no cover
                "axes cannot be None for operator Unsqueeze (Unsqueeze_13).")
        return (sq, )

    def _infer_shapes(self, x, axes=None):  # pylint: disable=W0221
        return (ShapeObject(None, dtype=x.dtype), )


if onnx_opset_version() >= 13:
    Unsqueeze = Unsqueeze_13
elif onnx_opset_version() >= 11:
    Unsqueeze = Unsqueeze_11
else:
    Unsqueeze = Unsqueeze_1



"""
@file
@brief Functions to show shortened options in :epkg:`asv` benchmarks.
"""


def expand_onnx_options(model, optim):
    """
    Expands shortened options. Long names hide some part
    of graphs in :epkg:`asv` benchmark. This trick converts
    a string into real conversions options.

    @param      model       model class (:epkg:`scikit-learn`)
    @param      optim       option
    @return                 expanded options

    It is the reverse of function @see fn shorten_onnx_options.
    The following options are handled:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from sklearn.linear_model import LogisticRegression
        from mlprodict.tools.asv_options_helper import expand_onnx_options

        for name in ['cdist', 'nozipmap', 'raw_scores']:
            print(name, ':', expand_onnx_options(LogisticRegression, name))
    """
    if optim == 'cdist':
        options = {model.__class__: {'optim': 'cdist'}}
    elif optim == 'nozipmap':
        options = {model.__class__: {'zipmap': False}}
    elif optim == 'raw_scores':
        options = {model.__class__: {'raw_scores': True, 'zipmap': False}}
    else:
        options = optim  # pragma: no cover
    return options


def shorten_onnx_options(model, opts):
    """
    Shortens onnx options into a string.
    Long names hide some part
    of graphs in :epkg:`asv` benchmark.

    @param      model       model class (:epkg:`scikit-learn`)
    @param      opts        options
    @return                 shortened options

    It is the reverse of function @see fn expand_onnx_options.
    """
    if opts is None:
        return opts
    if opts == {model: {'optim': 'cdist'}}:
        return 'cdist'
    if opts == {model: {'zipmap': False}}:
        return 'nozipmap'
    if opts == {model: {'raw_scores': True, 'zipmap': False}}:
        return 'raw_scores'
    return None


def benchmark_version():
    """
    Returns the list of ONNX version to benchmarks.
    Following snippet of code shows which version is
    current done.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.tools.asv_options_helper import benchmark_version
        print(benchmark_version())
    """
    return [13]


def ir_version():
    """
    Returns the preferred `IR_VERSION
    <https://github.com/onnx/onnx/blob/master/docs/IR.md#onnx-versioning>`_.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.tools.asv_options_helper import ir_version
        print(ir_version())
    """
    return [7]


def get_opset_number_from_onnx(benchmark=True):
    """
    Retuns the current :epkg:`onnx` opset
    based on the installed version of :epkg:`onnx`.

    @param      benchmark       returns the latest
                                version usable for benchmark
    @eturn                      opset number
    """
    if benchmark:
        return benchmark_version()[-1]
    from onnx.defs import onnx_opset_version  # pylint: disable=W0611
    return onnx_opset_version()


def get_ir_version_from_onnx(benchmark=True):
    """
    Retuns the current :epkg:`onnx` :epkg:`IR_VERSION`
    based on the installed version of :epkg:`onnx`.

    @param      benchmark       returns the latest
                                version usable for benchmark
    @eturn                      opset number

    .. faqref::
        :title: Failed to load model with error: Unknown model file format version.
        :lid: l-onnx-ir-version-fail

        :epkg:`onnxruntime` (or ``runtime='onnxruntime1'`` with @see cl OnnxInference)
        fails sometimes to load a model showing the following error messsage:

        ::

            RuntimeError: Unable to create InferenceSession due to '[ONNXRuntimeError] :
            2 : INVALID_ARGUMENT : Failed to load model with error: Unknown model file format version.'

        This case is due to metadata ``ir_version`` which defines the
        :epkg:`IR_VERSION` or *ONNX version*. When a model is machine learned
        model is converted, it is usually done with the default version
        (``ir_version``) returned by the :epkg:`onnx` package.
        :epkg:`onnxruntime` raises the above mentioned error message
        when this version (``ir_version``) is too recent. In this case,
        :epkg:`onnxruntime` should be updated to the latest version
        available or the metadata ``ir_version`` can just be changed to
        a lower number. Th function @see fn get_ir_version_from_onnx
        returns the latest tested version with *mlprodict*.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning

            from sklearn.linear_model import LinearRegression
            from sklearn.datasets import load_iris
            from mlprodict.onnxrt import OnnxInference
            import numpy

            iris = load_iris()
            X = iris.data[:, :2]
            y = iris.target
            lr = LinearRegression()
            lr.fit(X, y)

            # Conversion into ONNX.
            from mlprodict.onnx_conv import to_onnx
            model_onnx = to_onnx(lr, X.astype(numpy.float32),
                                 target_opset=12)
            print("ir_version", model_onnx.ir_version)

            # Change ir_version
            model_onnx.ir_version = 6

            # Predictions with onnxruntime
            oinf = OnnxInference(model_onnx, runtime='onnxruntime1')
            ypred = oinf.run({'X': X[:5].astype(numpy.float32)})
            print("ONNX output:", ypred)

            # To avoid keep a fixed version number, you can use
            # the value returned by function get_ir_version_from_onnx
            from mlprodict.tools import get_ir_version_from_onnx
            model_onnx.ir_version = get_ir_version_from_onnx()
            print("ir_version", model_onnx.ir_version)
    """
    if benchmark:
        return ir_version()[-1]
    from onnx import IR_VERSION  # pylint: disable=W0611
    return IR_VERSION


def display_onnx(model_onnx, max_length=1000):
    """
    Returns a shortened string of the model.

    @param      model_onnx      onnx model
    @param      max_length      maximal string length
    @return                     string
    """
    res = str(model_onnx)
    if max_length is None or len(res) <= max_length:
        return res
    begin = res[:max_length // 2]
    end = res[-max_length // 2:]
    return "\n".join([begin, '[...]', end])


def version2number(vers):
    """
    Converts a version number into a real number.
    """
    spl = vers.split('.')
    r = 0
    for i, s in enumerate(spl):
        try:
            vi = int(s)
        except ValueError:
            vi = 0
        r += vi * 10 ** (-i * 3)
    return r



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRun


class Sum(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, *args):  # pylint: disable=W0221
        return (sum(args), )

    def _infer_shapes(self, *args):  # pylint: disable=W0221
        return (args[0], )

    def to_python(self, inputs):
        return None, "return sum([%s])" % ", ".join(inputs)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRunUnaryNum


class Identity(OpRunUnaryNum):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               **options)

    def _run(self, a):  # pylint: disable=W0221
        if self.inplaces.get(0, False):
            return (a, )
        return (a.copy(), )

    def to_python(self, inputs):
        return "", "return %s.copy()" % inputs[0]



"""
@file
@brief Optimisations of :epkg:`ONNX` graphs.
"""
from ._onnx_optimisation_common import _apply_optimisation_on_graph
from .onnx_optimisation_identity import onnx_remove_node_identity
from .onnx_optimisation_redundant import onnx_remove_node_redundant
from .onnx_optimisation_unused import onnx_remove_node_unused


def onnx_remove_node(onnx_model, recursive=True, debug_info=None, **options):
    """
    Removes as many nodes as possible without changing
    the outcome. It applies @see fn onnx_remove_node_identity,
    then @see fn onnx_remove_node_redundant.

    @param      onnx_model      onnx model
    @param      recursive       looks into subgraphs
    @param      debug_info      debug information (private)
    @param      options         additional options
    @return                     new onnx _model
    """
    if debug_info is None:
        debug_info = [str(type(onnx_model)).split('.')[-1].strip("'>")]
    else:
        debug_info = debug_info + \
            [str(type(onnx_model)).split('.')[-1].strip("'>")]

    if hasattr(onnx_model, 'graph'):
        return _apply_optimisation_on_graph(
            onnx_remove_node, onnx_model,
            recursive=recursive, debug_info=debug_info,
            **options)

    graph = onnx_model
    graph = onnx_remove_node_unused(
        graph, recursive=recursive, debug_info=debug_info, **options)
    graph = onnx_remove_node_identity(
        graph, recursive=recursive, debug_info=debug_info, **options)
    graph = onnx_remove_node_redundant(
        graph, recursive=recursive, debug_info=debug_info, **options)
    return graph



"""
@file
@brief Intermediate class between :epkg:`numpy` and :epkg:`onnx`.

.. versionadded:: 0.6
"""
import numpy
from onnx.helper import make_tensor
from skl2onnx.common.data_types import guess_numpy_type
from skl2onnx.common._topology import Variable  # pylint: disable=E0611,E0001
from skl2onnx.algebra.onnx_ops import (  # pylint: disable=E0611
    OnnxAdd, OnnxAnd,
    OnnxCast, OnnxConstantOfShape,
    OnnxDiv,
    OnnxEqual,
    OnnxFlatten,
    OnnxGather, OnnxGreater, OnnxGreaterOrEqual,
    OnnxIdentity,
    OnnxLess, OnnxLessOrEqual,
    OnnxMatMul, OnnxMod, OnnxMul,
    OnnxNeg, OnnxNot,
    OnnxOr,
    OnnxPow,
    OnnxReduceSum, OnnxReshape,
    OnnxScatterElements, OnnxShape, OnnxSize, OnnxSlice,
    OnnxSqueeze, OnnxSub,
    OnnxTopK, OnnxTranspose
)
from skl2onnx.algebra.onnx_operator import OnnxOperatorItem
from skl2onnx.common.data_types import _guess_numpy_type
from ..tools.onnx2py_helper import guess_proto_dtype


try:
    numpy_bool = numpy.bool_
except AttributeError:  # pragma: no cover
    numpy_bool = bool
try:
    numpy_str = numpy.str
except AttributeError:  # pragma: no cover
    numpy_str = str


class OnnxVar:
    """
    Variables used into :epkg:`onnx` computation.

    :param inputs: variable name or object
    :param op: :epkg:`ONNX` operator
    :param select_output: if multiple output are returned by
        ONNX operator *op*, it takes only one specifed by this
        argument
    :param dtype: specifies the type of the variable
        held by this class (*op* is None) in that case
    :param kwargs: addition argument to give operator *op*

    .. versionadded:: 0.6
    """

    def __init__(self, *inputs, op=None, select_output=None,
                 dtype=None, **kwargs):
        self.inputs = inputs
        self.select_output = select_output
        self.onnx_op = op
        self.alg_ = None
        self.onnx_op_kwargs = kwargs
        if dtype is not None and (op is not None or len(inputs) != 1):
            raise RuntimeError(  # pragma: no cover
                "dtype can only be used if op is None or len(inputs) == 1.")
        for i, inp in enumerate(self.inputs):
            if isinstance(inp, type):
                raise TypeError(  # pragma: no cover
                    "Unexpected type for input %d - %r." % (i, inp))
        self.dtype = self._guess_dtype(dtype)

    def _guess_dtype(self, dtype):
        "Guesses dtype when not specified."
        if dtype is not None:
            return dtype
        dtypes = []
        for i, inp in enumerate(self.inputs):
            if isinstance(inp, str):
                return None
            if isinstance(inp, numpy.ndarray):
                dtypes.append(inp.dtype)
            elif isinstance(inp, Variable):
                dt = guess_numpy_type(inp.type)
                dtypes.append(dt)
            elif isinstance(inp, OnnxVar):
                dtypes.append(inp.dtype)
            elif isinstance(inp, MultiOnnxVar):
                dtypes.append(inp._guess_dtype(dtype))
            elif isinstance(inp, (numpy.float32, numpy.float64, numpy.int32,
                                  numpy.int64)):
                dtypes.append(inp.dtype)
            elif isinstance(inp, numpy_str):
                dtypes.append(numpy_str)
            elif isinstance(inp, numpy_bool):
                dtypes.append(numpy_bool)
            elif isinstance(inp, int):
                dtypes.append(numpy.int64)  # pragma: no cover
            elif isinstance(inp, float):
                dtypes.append(numpy.float64)
            elif hasattr(inp, 'fit'):
                # scikit-learn model
                continue
            else:
                raise TypeError(
                    "Unexpected type for input %i type=%r." % (i, type(inp)))
        dtypes = [_ for _ in dtypes if _ is not None]
        unique = set(dtypes)
        if len(unique) != 1:
            return None
        return dtypes[0]

    def __repr__(self):
        "usual"
        args = []
        for inp in self.inputs:
            args.append(repr(inp))
        if self.onnx_op is not None:
            if isinstance(self.onnx_op, str):
                args.append("op=%r" % self.onnx_op)
            else:
                args.append("op=%s" % self.onnx_op.__name__)
        if self.select_output is not None:
            args.append("select_output=%r" % self.select_output)
        if self.dtype is not None and self.dtype != self._guess_dtype(None):
            args.append("dtype=%r" % self.dtype)
        for k, v in sorted(self.onnx_op_kwargs.items()):
            args.append("%s=%r" % (k, v))
        res = "%s(%s)" % (self.__class__.__name__, ", ".join(args))
        return res

    def to_algebra(self, op_version=None):
        """
        Converts the variable into an operator.
        """
        if self.alg_ is None:
            if self.onnx_op is None:
                if len(self.inputs) != 1:
                    raise RuntimeError(  # pragma: no cover
                        "Unexpected number of inputs, 1 expected, "
                        "got {} instead.".format(self.inputs))
                if self.dtype is None or hasattr(self.inputs[0], 'onnx_name'):
                    self.alg_ = self.inputs[0]
                else:
                    self.alg_ = (
                        self.inputs[0], _guess_numpy_type(self.dtype, None))
            else:
                if isinstance(self.onnx_op, str):
                    var = self._custom_op(*self.inputs, op_version=op_version,
                                          **self.onnx_op_kwargs)
                    alg = var.to_algebra(op_version=op_version)
                    if not hasattr(self, 'alg_'):
                        raise RuntimeError(  # pragma: no cover
                            "Missing attribute 'alg_'.")
                    self.alg_ = alg
                    return alg

                new_inputs = []
                for inp in self.inputs:
                    if hasattr(inp, 'fit'):
                        # scikit-learn model
                        new_inputs.append(inp)
                    elif isinstance(inp, (
                            int, float, str, numpy.ndarray, numpy.int32,
                            numpy.int64, numpy.float32, numpy.float64,
                            numpy_bool, numpy_str, numpy.int8, numpy.uint8,
                            numpy.int16, numpy.uint16, numpy.uint32, numpy.uint64)):
                        new_inputs.append(inp)
                    else:
                        new_inputs.append(
                            inp.to_algebra(op_version=op_version))

                res = self.onnx_op(*new_inputs, op_version=op_version,
                                   **self.onnx_op_kwargs)
                if self.select_output is None:
                    self.alg_ = res
                else:
                    self.alg_ = res[self.select_output]
        return self.alg_

    def _custom_op(self, *args, op_version=None, runtime=None, **kwargs):
        """
        This could be handled before a call to this method
        but this method can change the conversion of an non-existing
        operator depending on the given opset.
        """
        if self.onnx_op == 'filter':
            return self._custom_op_filter(*args, op_version=op_version,
                                          runtime=runtime, **kwargs)
        raise NotImplementedError(  # pragma: no cover
            "Unexpected custom operator %r." % self.onnx_op)

    def _custom_op_filter(self, *args, op_version=None, runtime=None, **kwargs):
        """
        This could be handled before a call to this method
        but this method can change the conversion of an non-existing
        operator depending on the given opset.
        """
        if len(args) != 2:
            raise RuntimeError(  # pragma: no cover
                "Custom op 'filter' expects two inputs not %r." % len(args))
        if len(kwargs) != 0:
            raise RuntimeError(  # pragma: no cover
                "Custom op 'filter' expects no arguments but got %r." % kwargs)
        mat, index = args
        cast = OnnxVar(index.astype(numpy.int64), op=OnnxSqueeze)
        n1 = OnnxVar(cast, op=OnnxReduceSum, keepdims=1)
        indices = OnnxVar(cast, n1, op=OnnxTopK, select_output=1)
        return OnnxVar(mat, indices, op=OnnxGather)

    @property
    def T(self):
        "Transpose."
        return OnnxVar(self, op=OnnxTranspose)

    def astype(self, dtype):
        "Cast"
        return OnnxVar(self, op=OnnxCast, to=guess_proto_dtype(dtype))

    @property
    def shape(self):
        "Shape"
        return OnnxVar(self, op=OnnxShape)

    @property
    def size(self):
        "Size"
        return OnnxVar(self, op=OnnxSize)

    def reshape(self, shape):
        "Reshape"
        if isinstance(shape, (tuple, list)):
            shape = numpy.array(shape, dtype=numpy.int64)
        return OnnxVar(self, shape, op=OnnxReshape)

    def _make_array(self, y):
        """Converts *y* into an array if not."""
        if hasattr(y, 'dtype') and not isinstance(y, (numpy.ndarray, OnnxVar)):
            return numpy.full((1, ), y, dtype=y.dtype)
        if isinstance(y, (float, int, str)):
            return numpy.array([y])
        return y

    def __add__(self, y):
        "Addition."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxAdd)

    def __sub__(self, y):
        "Subtraction."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxSub)

    def __mul__(self, y):
        "Multiplication."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxMul)

    def __pow__(self, y):
        "Power."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxPow)

    def __mod__(self, y):
        "Modulo."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxMod)

    def __matmul__(self, y):
        "Matrix multiplication."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxMatMul)

    def __truediv__(self, y):
        "Division, no difference between `/` and `//`."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxDiv)

    def __floordiv__(self, y):
        "Division, no difference between `/` and `//`."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxDiv)

    def __eq__(self, y):
        "Equality."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxEqual)

    def __ne__(self, y):
        "Difference."
        y = self._make_array(y)
        return OnnxVar(OnnxVar(self, y, op=OnnxEqual), op=OnnxNot)

    def __ge__(self, y):
        "Greater or Equal."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxGreaterOrEqual)

    def __gt__(self, y):
        "Greater."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxGreater)

    def __le__(self, y):
        "Less or Equal."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxLessOrEqual)

    def __lt__(self, y):
        "Less."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxLess)

    def __and__(self, y):
        "And."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxAnd)

    def __or__(self, y):
        "And."
        y = self._make_array(y)
        return OnnxVar(self, y, op=OnnxOr)

    def not_(self):
        "Not."
        return OnnxVar(self, op=OnnxNot)

    def __neg__(self):
        "Neg."
        return OnnxVar(self, op=OnnxNeg)

    def __getitem__(self, index):
        """
        Deals with multiple scenarios.
        * *index* is an integer or a slice, a tuple of integers and slices,
          example: `[0, 1]`, `[:5, :6]`, `[::2]` (**scenario 1**)
        * *index* is an *ONNX* object (more precisely an instance of
          @see cl OnnxVar), then the method assumes it is an array of
          boolean to select a subset of the tensor along the first axis,
          example: `mat[mat == 0]` (**scenario 2**)
        """
        if isinstance(index, OnnxVar):
            # scenario 2
            return OnnxVar(self, index, op='filter')

        if not isinstance(index, tuple):
            index = (index, )

        # scenario 1
        starts = []
        ends = []
        axes = []
        steps = []
        axis_squeeze = []
        for i, ind in enumerate(index):
            if isinstance(ind, int):
                starts.append(ind)
                ends.append(ind + 1)
                axes.append(i)
                steps.append(1)
                axis_squeeze.append(i)
                continue
            if isinstance(ind, slice):
                if ind.start is None and ind.stop is None and ind.step is None:
                    continue
                start = 0 if ind.start is None else ind.start
                end = -1 if ind.stop is None else ind.stop
                step = 1 if ind.step is None else ind.step
                starts.append(start)
                ends.append(end)
                axes.append(i)
                steps.append(step)
                continue
            raise NotImplementedError(  # pragma: no cover
                "Not implemented for type %r." % type(ind))
        if max(steps) == min(steps) == 1:
            steps = None
        else:
            steps = numpy.array(steps, dtype=numpy.int64)
        starts = numpy.array(starts, dtype=numpy.int64)
        ends = numpy.array(ends, dtype=numpy.int64)
        axes = numpy.array(axes, dtype=numpy.int64)
        if steps is None:
            sliced = OnnxVar(self, starts, ends, axes, op=OnnxSlice)
        else:
            sliced = OnnxVar(self, starts, ends, axes, steps, op=OnnxSlice)
        if len(axis_squeeze) > 0:
            return OnnxVar(
                sliced, numpy.array(axis_squeeze, dtype=numpy.int64),
                op=OnnxSqueeze)
        return sliced

    def __setitem__(self, index, value):
        """
        Only supports vectors (1D tensor).
        * *index* is an integer or a slice, a tuple of integers and slices,
          example: `[0]`, `[:5]`, `[::2]` (**scenario 1**)
        * *index* is an *ONNX* object (more precisely an instance of
          @see cl OnnxVar), then the method assumes it is an array of
          boolean to select a subset of the tensor along the first axis,
          example: `mat[mat == 0]` (**scenario 2**)
        This processing is applied before the operator it contains.
        A copy should be made (Identity node or copy method).
        """
        if self.onnx_op is not None and self.onnx_op is not OnnxIdentity:
            raise RuntimeError(  # pragma: no cover
                "A copy should be made before setting new values on a matrix. "
                "Method copy() would do that.")

        if isinstance(index, OnnxVar):
            # scenario 2
            raise NotImplementedError()  # pragma: no cover

        if not isinstance(index, tuple):
            index = (index, )

        # scenario 1
        if len(index) == 1:
            return self._setitem1i_(index[0], value)
        raise NotImplementedError(  # pragma: no cover
            "Indices in %d dimensions are not implemented yet." % len(index))

    def _setitem1i_(self, index, value):
        sl = None
        if isinstance(index, slice):
            start = 0 if index.start is None else index.start
            stop = index.stop
            step = index.step
            sl = [start, stop, step]
        elif isinstance(index, int):
            sl = [index, index + 1, 1]
        else:
            raise NotImplementedError(  # pragma: no cover
                "Unable to assign new values due to unexpected type %r."
                "" % type(index))

        if sl[1] is None and isinstance(value, numpy.ndarray):
            sl[1] = sl[0] + value.size
        if sl[1] is None:
            if sl[2] is not None and sl[2] != 1:
                raise NotImplementedError(  # pragma: no cover
                    "If the length is not known, step must be 1 not %d." % sl[2])
            value = make_tensor(
                "value", guess_proto_dtype(value.dtype), (1, ), [value])  # pylint: disable=E1101
            inp = self.inputs[0]
            if not isinstance(inp, OnnxVar):
                raise RuntimeError(  # pragma: no cover
                    "Input must be an instance of OnnxVar not %r." % type(inp))
            cst = OnnxVar(inp.shape, op=OnnxConstantOfShape, value=value)
            ext = inp[:sl[0]]
            indices = numpy.arange(0, sl[0]).astype(numpy.int64)
            add_step = OnnxVar(cst, indices, ext,
                               op=OnnxScatterElements, axis=0)
        else:
            indices = numpy.arange(sl[0], sl[1], sl[2]).astype(numpy.int64)
            if isinstance(value, numpy.ndarray):
                values = value
            else:
                values = numpy.full(indices.shape, value)
            add_step = OnnxVar(self.inputs[0], indices, values,
                               op=OnnxScatterElements, axis=0)

        self.inputs = [add_step]
        return self

    def copy(self):
        """
        Returns a copy of self (use of Identity node).
        """
        return OnnxVar(self, op=OnnxIdentity)

    def flatten(self, axis=0):
        """
        Flattens a matrix (see :epkg:`numpy:ndarray:flatten`).

        :param axis: only flatten from axis to the end.
        :return: @see cl OnnxVariable
        """
        fl = OnnxVar(self, op=OnnxFlatten, axis=axis)
        if axis == 0:
            return OnnxVar(fl, numpy.array([0], dtype=numpy.int64),
                           op=OnnxSqueeze)
        return fl


class TupleOnnxAny:
    """
    Class used to return multiple @see cl OnnxVar
    at the same time.
    """

    def __init__(self, first, *args):
        if isinstance(first, (list, tuple)):
            raise TypeError(  # pragma: no cover
                "Unexpected type for first %r." % type(first))
        if len(args) > 0:
            self.values = (first,) + args
            self.unique = None
        else:
            self.values = None
            self.unique = first
        if self.values is not None and self.unique is not None:
            raise RuntimeError(  # pragma: no cover
                "Unexpected configuration. One member (values or unique) must be "
                "null, unique=%r, values=%r" % (self.unique, self.values))
        if self.values is None and self.unique is None:
            raise RuntimeError(  # pragma: no cover
                "Unexpected configuration. One member (values or unique) must be "
                "not null.")

    def __len__(self):
        "usual"
        if self.values is None:
            raise NotImplementedError(  # pragma: no cover
                "Not yet implemented in this case unique=%r, "
                "values=%r." % (self.unique, self.values))
        return len(self.values)

    def __iter__(self):
        "Iterates on the outputs."
        if self.values is None:
            raise NotImplementedError(  # pragma: no cover
                "Not yet implemented in this case.")
        for v in self.values:
            yield v

    def __getitem__(self, i):
        "usual"
        if self.values is None:
            return self.unique[i]
        return self.values[i]

    def get_output_type_inference(self, input_shapes=None):
        """
        Returns the expected output types in a list.
        """
        if self.values is None:
            if hasattr(self.unique, 'get_output_type_inference'):
                return self.unique.get_output_type_inference(input_shapes)
        raise NotImplementedError(  # pragma: no cover
            "Not implemented yet unique=%r values=%r." % (
                self.unique, self.values))

    @property
    def outputs(self):
        "Returns 'output_names' of attribute 'unique'."
        if self.values is None:
            if hasattr(self.unique, 'to_onnx'):
                return self.unique.outputs
        raise NotImplementedError(  # pragma: no cover
            "Not implemented yet unique=%r values=%r." % (
                self.unique, self.values))

    @property
    def output_names(self):
        "Returns 'output_names' of attribute 'unique'."
        if self.values is None:
            if hasattr(self.unique, 'to_onnx'):
                return self.unique.output_names
        raise NotImplementedError(  # pragma: no cover
            "Not implemented yet unique=%r values=%r." % (
                self.unique, self.values))

    @output_names.setter
    def output_names(self, value):
        """
        Updates 'output_names' of attribute 'unique'
        or every output name of attribute 'values'.
        """
        if self.values is None:
            if (hasattr(self.unique, 'to_onnx') or
                    hasattr(self.unique, 'add_to')):
                if len(value) > 1:
                    self.values = tuple(
                        OnnxIdentity(self.unique[i], output_names=value[i:i + 1],
                                     op_version=self.unique.op_version)
                        for i in range(0, len(value)))
                    self.unique = None
                    return
                self.unique.output_names = value
                return
            raise NotImplementedError(  # pragma: no cover
                "Not implemented yet, value=%r, unique=%r values=%r." % (
                    value, self.unique, self.values))
        if self.values is not None and len(self.values) == len(value):
            for name, v in zip(value, self.values):
                v.output_names = [name]
            return
        raise NotImplementedError(  # pragma: no cover
            "Not implemented yet, value=%r, unique=%r values=%r." % (
                value, self.unique, self.values))

    def add_to(self, scope, container, operator=None, run_converters=False):
        """
        Adds outputs to the container if not already added,
        registered the outputs if the node is not final.

        :param scope: scope
        :param container: container
        :param operator: overwrite inputs
        :param run_converters: must be True if called from method `to_onnx`
        """
        if self.values is not None:
            for v in self.values:
                v.add_to(scope, container, operator=operator,
                         run_converters=run_converters)
            return
        if self.unique is not None:
            self.unique.add_to(scope, container, operator=operator,
                               run_converters=run_converters)
            return
        raise RuntimeError(  # pragma: no cover
            "Attributes 'unique' and 'values' cannot be both null.")

    def to_onnx(self, *args, **kwargs):  # pylint: disable=W0222
        "Converts the underlying class into an ONNX graph."
        if self.values is None:
            if hasattr(self.unique, 'to_onnx'):
                return self.unique.to_onnx(*args, **kwargs)
            raise NotImplementedError(  # pragma: no cover
                "Not implemented yet unique=%r values=%r args=%r "
                "kwargs=%r." % (self.unique, self.values, args, kwargs))
        if self.values is not None:
            if len(self.values) == len(kwargs.get('outputs', [])):
                return self.values[0].to_onnx(
                    *args, other_outputs=self.values[1:], **kwargs)
        raise NotImplementedError(  # pragma: no cover
            "Not implemented yet unique=%r values=%r args=%r "
            "kwargs=%r." % (self.unique, self.values, args, kwargs))


class MultiOnnxVar:
    """
    Class used to return multiple @see cl OnnxVar
    at the same time.
    """

    def __init__(self, *inputs, op=None, dtype=None, **kwargs):
        "constructor"
        self.onxvar = OnnxVar(*inputs, op=op, dtype=None, **kwargs)
        self.alg_ = None

    def _guess_dtype(self, dtype):
        "Guesses dtype when not specified."
        return self.onxvar._guess_dtype(dtype)

    @property
    def inputs(self):
        "Returns `self.onxvar.inputs`."
        return self.onxvar.inputs

    @property
    def onnx_op(self):
        "Returns `self.onxvar.onnx_op`."
        return self.onxvar.onnx_op

    @property
    def onnx_op_kwargs(self):
        "Returns `self.onxvar.onnx_op_kwargs`."
        return self.onxvar.onnx_op_kwargs

    def to_algebra(self, op_version=None):
        """
        Converts the variable into an operator.
        """
        if self.alg_ is None:
            new_inputs = []
            for inp in self.inputs:
                if isinstance(inp, (
                        int, float, str, numpy.ndarray, numpy.int32,
                        numpy.int64, numpy.float32, numpy.float64,
                        numpy_bool, numpy_str, numpy.int8, numpy.uint8,
                        numpy.int16, numpy.uint16, numpy.uint32, numpy.uint64)):
                    new_inputs.append(inp)
                elif hasattr(inp, 'fit'):
                    # scikit-learn models
                    new_inputs.append(inp)
                else:
                    new_inputs.append(
                        inp.to_algebra(op_version=op_version))

            if self.onnx_op is None:
                if len(new_inputs) == 1:
                    self.alg_ = TupleOnnxAny(new_inputs[0])
                else:
                    self.alg_ = TupleOnnxAny(new_inputs[0], *(new_inputs[1:]))
            else:
                res = self.onnx_op(  # pylint: disable=E1102
                    *new_inputs, op_version=op_version, **self.onnx_op_kwargs)
                self.alg_ = TupleOnnxAny(res)
        return self.alg_

    def __getitem__(self, index):
        """
        Returns the ith elements.
        """
        return OnnxVar(self, index=index, op=OnnxOperatorItem)



"""
@file
@brief Converters for models from :epkg:`mlinsights`.
"""
from sklearn.base import ClassifierMixin
from skl2onnx import get_model_alias
from skl2onnx.common.data_types import FloatTensorType
from skl2onnx.common._registration import get_shape_calculator
from skl2onnx._parse import parse_sklearn
from skl2onnx.common._apply_operation import apply_identity


def _model_outputs(existing_scope, model, inputs, custom_parsers=None):
    """
    Retrieves the outputs of one particular models.
    """
    scope = existing_scope.temp()
    if custom_parsers is not None and model in custom_parsers:
        return custom_parsers[model](
            scope, model, inputs, custom_parsers=custom_parsers)
    return parse_sklearn(scope, model, inputs, custom_parsers=custom_parsers)


def parser_transfer_transformer(scope, model, inputs, custom_parsers=None):
    """
    Parser for :epkg:`TransferTransformer`.
    """
    if custom_parsers is not None and model in custom_parsers:
        return custom_parsers[model](
            scope, model, inputs, custom_parsers=custom_parsers)

    if model.method == 'predict_proba':
        name = 'probabilities'
    elif model.method == 'transform':
        name = 'variable'
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unable to defined the output for method='{}' and model='{}'.".format(
                model.method, model.__class__.__name__))

    prob = scope.declare_local_variable(name, FloatTensorType())
    alias = get_model_alias(type(model))
    this_operator = scope.declare_local_operator(alias, model)
    this_operator.inputs = inputs
    this_operator.outputs.append(prob)
    return this_operator.outputs


def shape_calculator_transfer_transformer(operator):
    """
    Shape calculator :epkg:`TransferTransformer`.
    """
    op = operator.raw_operator
    alias = get_model_alias(type(op.estimator_))
    calc = get_shape_calculator(alias)

    scope = operator.scope_inst.temp()
    this_operator = scope.declare_local_operator(alias)
    this_operator.raw_operator = op.estimator_
    this_operator.inputs = operator.inputs
    res = _model_outputs(scope, op.estimator_, operator.inputs)
    this_operator.outputs.extend([
        scope.declare_local_variable(
            "%sTTS" % r.onnx_name, r.type) for r in res])
    this_operator.outputs = res
    calc(this_operator)

    if op.method == 'predict_proba':
        operator.outputs[0].type = this_operator.outputs[1].type
    elif op.method == 'transform':
        operator.outputs[0].type = this_operator.outputs[0].type
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unable to defined the output for method='{}' and model='{}'.".format(
                op.method, op.__class__.__name__))


def convert_transfer_transformer(scope, operator, container):
    """
    Converters for :epkg:`TransferTransformer`.
    """
    op = operator.raw_operator
    op_type = get_model_alias(type(op.estimator_))

    this_operator = scope.declare_local_operator(op_type)
    this_operator.raw_operator = op.estimator_
    this_operator.inputs = operator.inputs

    if isinstance(op.estimator_, ClassifierMixin):
        container.add_options(id(op.estimator_), {'zipmap': False})

    res = _model_outputs(scope.temp(), op.estimator_, operator.inputs)
    this_operator.outputs.extend([
        scope.declare_local_variable(
            "%sTTC" % r.onnx_name, r.type) for r in res])

    if op.method == 'predict_proba':
        index = 1
    elif op.method == 'transform':
        index = 0
    else:
        raise NotImplementedError(  # pragma: no cover
            "Unable to defined the output for method='{}' and model='{}'.".format(
                op.method, op.__class__.__name__))

    apply_identity(scope, this_operator.outputs[index].onnx_name,
                   operator.outputs[0].full_name, container,
                   operator_name=scope.get_unique_operator_name("IdentityTT"))



"""
@file Functions to creates a benchmark based on :epkg:`asv`
for many regressors and classifiers.
"""
import os
import sys
import json
import textwrap
import warnings
import re
from pyquickhelper.pycode.code_helper import remove_extra_spaces_and_pep8
try:
    from ._create_asv_helper import (
        default_asv_conf,
        flask_helper,
        pyspy_template,
        _handle_init_files,
        _asv_class_name,
        _read_patterns,
        _select_pattern_problem,
        _display_code_lines,
        add_model_import_init,
        find_missing_sklearn_imports)
except ImportError:  # pragma: no cover
    from mlprodict.asv_benchmark._create_asv_helper import (
        default_asv_conf,
        flask_helper,
        pyspy_template,
        _handle_init_files,
        _asv_class_name,
        _read_patterns,
        _select_pattern_problem,
        _display_code_lines,
        add_model_import_init,
        find_missing_sklearn_imports)

try:
    from ..tools.asv_options_helper import (
        get_opset_number_from_onnx, shorten_onnx_options)
    from ..onnxrt.validate.validate_helper import sklearn_operators
    from ..onnxrt.validate.validate import (
        _retrieve_problems_extra, _get_problem_data, _merge_options)
except (ValueError, ImportError):  # pragma: no cover
    from mlprodict.tools.asv_options_helper import get_opset_number_from_onnx
    from mlprodict.onnxrt.validate.validate_helper import sklearn_operators
    from mlprodict.onnxrt.validate.validate import (
        _retrieve_problems_extra, _get_problem_data, _merge_options)
    from mlprodict.tools.asv_options_helper import shorten_onnx_options
try:
    from ..testing.verify_code import verify_code
except (ValueError, ImportError):  # pragma: no cover
    from mlprodict.testing.verify_code import verify_code

# exec function does not import models but potentially
# requires all specific models used to define scenarios
try:
    from ..onnxrt.validate.validate_scenarios import *  # pylint: disable=W0614,W0401
except (ValueError, ImportError):  # pragma: no cover
    # Skips this step if used in a benchmark.
    pass


def create_asv_benchmark(
        location, opset_min=-1, opset_max=None,
        runtime=('scikit-learn', 'python_compiled'), models=None,
        skip_models=None, extended_list=True,
        dims=(1, 10, 100, 10000),
        n_features=(4, 20), dtype=None,
        verbose=0, fLOG=print, clean=True,
        conf_params=None, filter_exp=None,
        filter_scenario=None, flat=False,
        exc=False, build=None, execute=False,
        add_pyspy=False, env=None,
        matrix=None):
    """
    Creates an :epkg:`asv` benchmark in a folder
    but does not run it.

    :param n_features: number of features to try
    :param dims: number of observations to try
    :param verbose: integer from 0 (None) to 2 (full verbose)
    :param opset_min: tries every conversion from this minimum opset,
        -1 to get the current opset defined by module :epkg:`onnx`
    :param opset_max: tries every conversion up to maximum opset,
        -1 to get the current opset defined by module :epkg:`onnx`
    :param runtime: runtime to check, *scikit-learn*, *python*,
        *python_compiled* compiles the graph structure
        and is more efficient when the number of observations is
        small, *onnxruntime1* to check :epkg:`onnxruntime`,
        *onnxruntime2* to check every ONNX node independently
        with onnxruntime, many runtime can be checked at the same time
        if the value is a comma separated list
    :param models: list of models to test or empty
        string to test them all
    :param skip_models: models to skip
    :param extended_list: extends the list of :epkg:`scikit-learn` converters
        with converters implemented in this module
    :param n_features: change the default number of features for
        a specific problem, it can also be a comma separated list
    :param dtype: '32' or '64' or None for both,
        limits the test to one specific number types
    :param fLOG: logging function
    :param clean: clean the folder first, otherwise overwrites the content
    :param conf_params: to overwrite some of the configuration parameters
    :param filter_exp: function which tells if the experiment must be run,
        None to run all, takes *model, problem* as an input
    :param filter_scenario: second function which tells if the experiment must be run,
        None to run all, takes *model, problem, scenario, extra*
        as an input
    :param flat: one folder for all files or subfolders
    :param exc: if False, raises warnings instead of exceptions
        whenever possible
    :param build: where to put the outputs
    :param execute: execute each script to make sure
        imports are correct
    :param add_pyspy: add an extra folder with code to profile
        each configuration
    :param env: None to use the default configuration or ``same`` to use
        the current one
    :param matrix: specifies versions for a module,
        example: ``{'onnxruntime': ['1.1.1', '1.1.2']}``,
        if a package name starts with `'~'`, the package is removed
    :return: created files

    The default configuration is the following:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import pprint
        from mlprodict.asv_benchmark.create_asv import default_asv_conf

        pprint.pprint(default_asv_conf)

    The benchmark does not seem to work well with setting
    ``-environment existing:same``. The publishing fails.
    """
    if opset_min == -1:
        opset_min = get_opset_number_from_onnx()
    if opset_max == -1:
        opset_max = get_opset_number_from_onnx()  # pragma: no cover
    if verbose > 0 and fLOG is not None:  # pragma: no cover
        fLOG("[create_asv_benchmark] opset in [{}, {}].".format(
            opset_min, opset_max))

    # creates the folder if it does not exist.
    if not os.path.exists(location):
        if verbose > 0 and fLOG is not None:  # pragma: no cover
            fLOG("[create_asv_benchmark] create folder '{}'.".format(location))
        os.makedirs(location)  # pragma: no cover

    location_test = os.path.join(location, 'benches')
    if not os.path.exists(location_test):
        if verbose > 0 and fLOG is not None:
            fLOG("[create_asv_benchmark] create folder '{}'.".format(location_test))
        os.mkdir(location_test)

    # Cleans the content of the folder
    created = []
    if clean:
        for name in os.listdir(location_test):
            full_name = os.path.join(location_test, name)  # pragma: no cover
            if os.path.isfile(full_name):  # pragma: no cover
                os.remove(full_name)

    # configuration
    conf = default_asv_conf.copy()
    if conf_params is not None:
        for k, v in conf_params.items():
            conf[k] = v
    if build is not None:
        for fi in ['env_dir', 'results_dir', 'html_dir']:  # pragma: no cover
            conf[fi] = os.path.join(build, conf[fi])
    if env == 'same':
        if matrix is not None:
            raise ValueError(  # pragma: no cover
                "Parameter matrix must be None if env is 'same'.")
        conf['pythons'] = ['same']
        conf['matrix'] = {}
    elif matrix is not None:
        drop_keys = set(p for p in matrix if p.startswith('~'))
        matrix = {k: v for k, v in matrix.items() if k not in drop_keys}
        conf['matrix'] = {k: v for k,
                          v in conf['matrix'].items() if k not in drop_keys}
        conf['matrix'].update(matrix)
    elif env is not None:
        raise ValueError(  # pragma: no cover
            "Unable to handle env='{}'.".format(env))
    dest = os.path.join(location, "asv.conf.json")
    created.append(dest)
    with open(dest, "w", encoding='utf-8') as f:
        json.dump(conf, f, indent=4)
    if verbose > 0 and fLOG is not None:
        fLOG("[create_asv_benchmark] create 'asv.conf.json'.")

    # __init__.py
    dest = os.path.join(location, "__init__.py")
    with open(dest, "w", encoding='utf-8') as f:
        pass
    created.append(dest)
    if verbose > 0 and fLOG is not None:
        fLOG("[create_asv_benchmark] create '__init__.py'.")
    dest = os.path.join(location_test, '__init__.py')
    with open(dest, "w", encoding='utf-8') as f:
        pass
    created.append(dest)
    if verbose > 0 and fLOG is not None:
        fLOG("[create_asv_benchmark] create 'benches/__init__.py'.")

    # flask_server
    tool_dir = os.path.join(location, 'tools')
    if not os.path.exists(tool_dir):
        os.mkdir(tool_dir)
    fl = os.path.join(tool_dir, 'flask_serve.py')
    with open(fl, "w", encoding='utf-8') as f:
        f.write(flask_helper)
    if verbose > 0 and fLOG is not None:
        fLOG("[create_asv_benchmark] create 'flask_serve.py'.")

    # command line
    if sys.platform.startswith("win"):
        run_bash = os.path.join(tool_dir, 'run_asv.bat')  # pragma: no cover
    else:
        run_bash = os.path.join(tool_dir, 'run_asv.sh')
    with open(run_bash, 'w') as f:
        f.write(textwrap.dedent("""
            echo --BENCHRUN--
            python -m asv run --show-stderr --config ./asv.conf.json
            echo --PUBLISH--
            python -m asv publish --config ./asv.conf.json -o ./html
            echo --CSV--
            python -m mlprodict asv2csv -f ./results -o ./data_bench.csv
            """))

    # pyspy
    if add_pyspy:
        dest_pyspy = os.path.join(location, 'pyspy')
        if not os.path.exists(dest_pyspy):
            os.mkdir(dest_pyspy)
    else:
        dest_pyspy = None

    if verbose > 0 and fLOG is not None:
        fLOG("[create_asv_benchmark] create all tests.")

    created.extend(list(_enumerate_asv_benchmark_all_models(
        location_test, opset_min=opset_min, opset_max=opset_max,
        runtime=runtime, models=models,
        skip_models=skip_models, extended_list=extended_list,
        n_features=n_features, dtype=dtype,
        verbose=verbose, filter_exp=filter_exp,
        filter_scenario=filter_scenario,
        dims=dims, exc=exc, flat=flat,
        fLOG=fLOG, execute=execute,
        dest_pyspy=dest_pyspy)))

    if verbose > 0 and fLOG is not None:
        fLOG("[create_asv_benchmark] done.")
    return created


def _enumerate_asv_benchmark_all_models(  # pylint: disable=R0914
        location, opset_min=10, opset_max=None,
        runtime=('scikit-learn', 'python'), models=None,
        skip_models=None, extended_list=True,
        n_features=None, dtype=None,
        verbose=0, filter_exp=None,
        dims=None, filter_scenario=None,
        exc=True, flat=False, execute=False,
        dest_pyspy=None, fLOG=print):
    """
    Loops over all possible models and fills a folder
    with benchmarks following :epkg:`asv` concepts.

    :param n_features: number of features to try
    :param dims: number of observations to try
    :param verbose: integer from 0 (None) to 2 (full verbose)
    :param opset_min: tries every conversion from this minimum opset
    :param opset_max: tries every conversion up to maximum opset
    :param runtime: runtime to check, *scikit-learn*, *python*,
        *onnxruntime1* to check :epkg:`onnxruntime`,
        *onnxruntime2* to check every ONNX node independently
        with onnxruntime, many runtime can be checked at the same time
        if the value is a comma separated list
    :param models: list of models to test or empty
        string to test them all
    :param skip_models: models to skip
    :param extended_list: extends the list of :epkg:`scikit-learn` converters
        with converters implemented in this module
    :param n_features: change the default number of features for
        a specific problem, it can also be a comma separated list
    :param dtype: '32' or '64' or None for both,
        limits the test to one specific number types
    :param fLOG: logging function
    :param filter_exp: function which tells if the experiment must be run,
        None to run all, takes *model, problem* as an input
    :param filter_scenario: second function which tells if the experiment must be run,
        None to run all, takes *model, problem, scenario, extra*
        as an input
    :param exc: if False, raises warnings instead of exceptions
        whenever possible
    :param flat: one folder for all files or subfolders
    :param execute: execute each script to make sure
        imports are correct
    :param dest_pyspy: add a file to profile the prediction
        function with :epkg:`pyspy`
    """

    ops = [_ for _ in sklearn_operators(extended=extended_list)]
    patterns = _read_patterns()

    if models is not None:
        if not all(map(lambda m: isinstance(m, str), models)):
            raise ValueError(
                "models must be a set of strings.")  # pragma: no cover
        ops_ = [_ for _ in ops if _['name'] in models]
        if len(ops) == 0:
            raise ValueError("Parameter models is wrong: {}\n{}".format(  # pragma: no cover
                models, ops[0]))
        ops = ops_
    if skip_models is not None:
        ops = [m for m in ops if m['name'] not in skip_models]

    if verbose > 0:

        def iterate():
            for i, row in enumerate(ops):  # pragma: no cover
                fLOG("{}/{} - {}".format(i + 1, len(ops), row))
                yield row

        if verbose >= 11:
            verbose -= 10  # pragma: no cover
            loop = iterate()  # pragma: no cover
        else:
            try:
                from tqdm import trange

                def iterate_tqdm():
                    with trange(len(ops)) as t:
                        for i in t:
                            row = ops[i]
                            disp = row['name'] + " " * (28 - len(row['name']))
                            t.set_description("%s" % disp)
                            yield row

                loop = iterate_tqdm()

            except ImportError:  # pragma: no cover
                loop = iterate()
    else:
        loop = ops

    if opset_max is None:
        opset_max = get_opset_number_from_onnx()
    opsets = list(range(opset_min, opset_max + 1))
    all_created = set()

    # loop on all models
    for row in loop:

        model = row['cl']

        problems, extras = _retrieve_problems_extra(
            model, verbose, fLOG, extended_list)
        if extras is None or problems is None:
            # Not tested yet.
            continue  # pragma: no cover

        # flat or not flat
        created, location_model, prefix_import, dest_pyspy_model = _handle_init_files(
            model, flat, location, verbose, dest_pyspy, fLOG)
        for init in created:
            yield init

        # loops on problems
        for prob in problems:
            if filter_exp is not None and not filter_exp(model, prob):
                continue

            (X_train, X_test, y_train,
             y_test, Xort_test,
             init_types, conv_options, method_name,
             output_index, dofit, predict_kwargs) = _get_problem_data(prob, None)

            for scenario_extra in extras:
                subset_problems = None
                optimisations = None
                new_conv_options = None

                if len(scenario_extra) > 2:
                    options = scenario_extra[2]
                    if isinstance(options, dict):
                        subset_problems = options.get('subset_problems', None)
                        optimisations = options.get('optim', None)
                        new_conv_options = options.get('conv_options', None)
                    else:
                        subset_problems = options

                if subset_problems and isinstance(subset_problems, (list, set)):
                    if prob not in subset_problems:
                        # Skips unrelated problem for a specific configuration.
                        continue
                elif subset_problems is not None:
                    raise RuntimeError(  # pragma: no cover
                        "subset_problems must be a set or a list not {}.".format(
                            subset_problems))

                scenario, extra = scenario_extra[:2]
                if optimisations is None:
                    optimisations = [None]
                if new_conv_options is None:
                    new_conv_options = [{}]

                if (filter_scenario is not None and
                        not filter_scenario(model, prob, scenario,
                                            extra, new_conv_options)):
                    continue  # pragma: no cover

                if verbose >= 3 and fLOG is not None:
                    fLOG("[create_asv_benchmark] model={} scenario={} optim={} extra={} dofit={} (problem={} method_name='{}')".format(
                        model.__name__, scenario, optimisations, extra, dofit, prob, method_name))
                created = _create_asv_benchmark_file(
                    location_model, opsets=opsets,
                    model=model, scenario=scenario, optimisations=optimisations,
                    extra=extra, dofit=dofit, problem=prob,
                    runtime=runtime, new_conv_options=new_conv_options,
                    X_train=X_train, X_test=X_test, y_train=y_train,
                    y_test=y_test, Xort_test=Xort_test,
                    init_types=init_types, conv_options=conv_options,
                    method_name=method_name, dims=dims, n_features=n_features,
                    output_index=output_index, predict_kwargs=predict_kwargs,
                    exc=exc, prefix_import=prefix_import,
                    execute=execute, location_pyspy=dest_pyspy_model,
                    patterns=patterns)
                for cr in created:
                    if cr in all_created:
                        raise RuntimeError(  # pragma: no cover
                            "File '{}' was already created.".format(cr))
                    all_created.add(cr)
                    if verbose > 1 and fLOG is not None:
                        fLOG("[create_asv_benchmark] add '{}'.".format(cr))
                    yield cr


def _create_asv_benchmark_file(  # pylint: disable=R0914
        location, model, scenario, optimisations, new_conv_options,
        extra, dofit, problem, runtime, X_train, X_test, y_train,
        y_test, Xort_test, init_types, conv_options,
        method_name, n_features, dims, opsets,
        output_index, predict_kwargs, prefix_import,
        exc, execute=False, location_pyspy=None, patterns=None):
    """
    Creates a benchmark file based in the information received
    through the argument. It uses template @see cl TemplateBenchmark.
    """
    if patterns is None:
        raise ValueError("Patterns list is empty.")  # pragma: no cover

    def format_conv_options(d_options, class_name):
        if d_options is None:
            return None
        res = {}
        for k, v in d_options.items():
            if isinstance(k, type):
                if "." + class_name + "'" in str(k):
                    res[class_name] = v
                    continue
                raise ValueError(  # pragma: no cover
                    "Class '{}', unable to format options {}".format(
                        class_name, d_options))
            res[k] = v
        return res

    def _nick_name_options(model, opts):
        # Shorten common onnx options, see _CommonAsvSklBenchmark._to_onnx.
        if opts is None:
            return opts  # pragma: no cover
        short_opts = shorten_onnx_options(model, opts)
        if short_opts is not None:
            return short_opts
        res = {}
        for k, v in opts.items():
            if hasattr(k, '__name__'):
                res["####" + k.__name__ + "####"] = v
            else:
                res[k] = v  # pragma: no cover
        return res

    def _make_simple_name(name):
        simple_name = name.replace("bench_", "").replace("_bench", "")
        simple_name = simple_name.replace("bench.", "").replace(".bench", "")
        simple_name = simple_name.replace(".", "-")
        repl = {'_': '', 'solverliblinear': 'liblinear'}
        for k, v in repl.items():
            simple_name = simple_name.replace(k, v)
        return simple_name

    def _optdict2string(opt):
        if isinstance(opt, str):
            return opt
        if isinstance(opt, list):
            raise TypeError(
                "Unable to process type %r." % type(opt))
        reps = {True: 1, False: 0, 'zipmap': 'zm',
                'optim': 'opt'}
        info = []
        for k, v in sorted(opt.items()):
            if isinstance(v, dict):
                v = _optdict2string(v)
            if k.startswith('####'):
                k = ''
            i = '{}{}'.format(reps.get(k, k), reps.get(v, v))
            info.append(i)
        return "-".join(info)

    runtimes_abb = {
        'scikit-learn': 'skl',
        'onnxruntime1': 'ort',
        'onnxruntime2': 'ort2',
        'python': 'pyrt',
        'python_compiled': 'pyrtc',
    }
    runtime = [runtimes_abb[k] for k in runtime]

    # Looping over configuration.
    names = []
    for optimisation in optimisations:
        merged_options = [_merge_options(nconv_options, conv_options)
                          for nconv_options in new_conv_options]

        nck_opts = [_nick_name_options(model, opts)
                    for opts in merged_options]
        try:
            name = _asv_class_name(
                model, scenario, optimisation, extra,
                dofit, conv_options, problem,
                shorten=True)
        except ValueError as e:  # pragma: no cover
            if exc:
                raise e
            warnings.warn(str(e))
            continue
        filename = name.replace(".", "_") + ".py"
        try:
            class_content = _select_pattern_problem(problem, patterns)
        except ValueError as e:
            if exc:
                raise e  # pragma: no cover
            warnings.warn(str(e))
            continue
        full_class_name = _asv_class_name(
            model, scenario, optimisation, extra,
            dofit, conv_options, problem,
            shorten=False)
        class_name = name.replace(
            "bench.", "").replace(".", "_") + "_bench"

        # n_features, N, runtimes
        rep = {
            "['skl', 'pyrtc', 'ort'],  # values for runtime": str(runtime),
            "[1, 10, 100, 1000, 10000],  # values for N": str(dims),
            "[4, 20],  # values for nf": str(n_features),
            "[get_opset_number_from_onnx()],  # values for opset": str(opsets),
            "['float', 'double'],  # values for dtype":
                "['float']" if '-64' not in problem else "['double']",
            "[None],  # values for optim": "%r" % nck_opts,
        }
        for k, v in rep.items():
            if k not in class_content:
                raise ValueError("Unable to find '{}'\n{}.".format(  # pragma: no cover
                    k, class_content))
            class_content = class_content.replace(k, v + ',')
        class_content = class_content.split(
            "def _create_model(self):")[0].strip("\n ")
        if "####" in class_content:
            class_content = class_content.replace(
                "'####", "").replace("####'", "")
        if "####" in class_content:
            raise RuntimeError(  # pragma: no cover
                "Substring '####' should not be part of the script for '{}'\n{}".format(
                    model.__name__, class_content))

        # Model setup
        class_content, atts = add_model_import_init(
            class_content, model, optimisation,
            extra, merged_options)
        class_content = class_content.replace(
            "class TemplateBenchmark",
            "class {}".format(class_name))

        # dtype, dofit
        atts.append("chk_method_name = %r" % method_name)
        atts.append("par_scenario = %r" % scenario)
        atts.append("par_problem = %r" % problem)
        atts.append("par_optimisation = %r" % optimisation)
        if not dofit:
            atts.append("par_dofit = False")
        if merged_options is not None and len(merged_options) > 0:
            atts.append("par_convopts = %r" % format_conv_options(
                conv_options, model.__name__))
        atts.append("par_full_test_name = %r" % full_class_name)

        simple_name = _make_simple_name(name)
        atts.append("benchmark_name = %r" % simple_name)
        atts.append("pretty_name = %r" % simple_name)

        if atts:
            class_content = class_content.replace(
                "# additional parameters",
                "\n    ".join(atts))
        if prefix_import != '.':
            class_content = class_content.replace(
                " from .", "from .{}".format(prefix_import))

        # Check compilation
        try:
            compile(class_content, filename, 'exec')
        except SyntaxError as e:  # pragma: no cover
            raise SyntaxError("Unable to compile model '{}'\n{}".format(
                model.__name__, class_content)) from e

        # Verifies missing imports.
        to_import, _ = verify_code(class_content, exc=False)
        try:
            miss = find_missing_sklearn_imports(to_import)
        except ValueError as e:  # pragma: no cover
            raise ValueError(
                "Unable to check import in script\n{}".format(
                    class_content)) from e
        class_content = class_content.replace(
            "#  __IMPORTS__", "\n".join(miss))
        verify_code(class_content, exc=True)
        class_content = class_content.replace(
            "par_extra = {", "par_extra = {\n")
        class_content = remove_extra_spaces_and_pep8(
            class_content, aggressive=True)

        # Check compilation again
        try:
            obj = compile(class_content, filename, 'exec')
        except SyntaxError as e:  # pragma: no cover
            raise SyntaxError("Unable to compile model '{}'\n{}".format(
                model.__name__,
                _display_code_lines(class_content))) from e

        # executes to check import
        if execute:
            try:
                exec(obj, globals(), locals())  # pylint: disable=W0122
            except Exception as e:  # pragma: no cover
                raise RuntimeError(
                    "Unable to process class '{}' ('{}') a script due to '{}'\n{}".format(
                        model.__name__, filename, str(e),
                        _display_code_lines(class_content))) from e

        # Saves
        fullname = os.path.join(location, filename)
        names.append(fullname)
        with open(fullname, "w", encoding='utf-8') as f:
            f.write(class_content)

        if location_pyspy is not None:
            # adding configuration for pyspy
            class_name = re.compile(
                'class ([A-Za-z_0-9]+)[(]').findall(class_content)[0]
            fullname_pyspy = os.path.splitext(
                os.path.join(location_pyspy, filename))[0]
            pyfold = os.path.splitext(os.path.split(fullname)[-1])[0]

            dtypes = ['float', 'double'] if '-64' in problem else ['float']
            for dim in dims:
                for nf in n_features:
                    for opset in opsets:
                        for dtype in dtypes:
                            for opt in nck_opts:
                                tmpl = pyspy_template.replace(
                                    '__PATH__', location)
                                tmpl = tmpl.replace(
                                    '__CLASSNAME__', class_name)
                                tmpl = tmpl.replace('__PYFOLD__', pyfold)
                                opt = "" if opt == {} else opt

                                first = True
                                for rt in runtime:
                                    if first:
                                        tmpl += textwrap.dedent("""

                                        def profile0_{rt}(iter, cl, N, nf, opset, dtype, optim):
                                            return setup_profile0(iter, cl, '{rt}', N, nf, opset, dtype, optim)
                                        iter = profile0_{rt}(iter, cl, {dim}, {nf}, {opset}, '{dtype}', {opt})
                                        print(datetime.now(), "iter", iter)

                                        """).format(rt=rt, dim=dim, nf=nf, opset=opset,
                                                    dtype=dtype, opt="%r" % opt)
                                        first = False

                                    tmpl += textwrap.dedent("""

                                    def profile_{rt}(iter, cl, N, nf, opset, dtype, optim):
                                        return setup_profile(iter, cl, '{rt}', N, nf, opset, dtype, optim)
                                    profile_{rt}(iter, cl, {dim}, {nf}, {opset}, '{dtype}', {opt})
                                    print(datetime.now(), "iter", iter)

                                    """).format(rt=rt, dim=dim, nf=nf, opset=opset,
                                                dtype=dtype, opt="%r" % opt)

                                thename = "{n}_{dim}_{nf}_{opset}_{dtype}_{opt}.py".format(
                                    n=fullname_pyspy, dim=dim, nf=nf,
                                    opset=opset, dtype=dtype, opt=_optdict2string(opt))
                                with open(thename, 'w', encoding='utf-8') as f:
                                    f.write(tmpl)
                                names.append(thename)

                                ext = '.bat' if sys.platform.startswith(
                                    'win') else '.sh'
                                script = os.path.splitext(thename)[0] + ext
                                short = os.path.splitext(
                                    os.path.split(thename)[-1])[0]
                                with open(script, 'w', encoding='utf-8') as f:
                                    f.write('py-spy record --native --function --rate=10 -o {n}_fct.svg -- {py} {n}.py\n'.format(
                                        py=sys.executable, n=short))
                                    f.write('py-spy record --native --rate=10 -o {n}_line.svg -- {py} {n}.py\n'.format(
                                        py=sys.executable, n=short))

    return names



"""
Common class for all benchmarks testing
converted models from :epkg:`scikit-learn`
with :epkg:`asv`. The benchmark can be run through
file :epkg:`run_asv.sh` on Linux or :epkg:`run_asv.bat` on
Windows.

.. warning::
    On Windows, you should avoid cloning the repository
    on a folder with a long full name. Visual Studio tends to
    abide by the rule of the maximum path length even though
    the system is told otherwise.
"""
import os
from datetime import datetime
import pickle
from logging import getLogger
import numpy
from sklearn import set_config
from sklearn.datasets import load_iris
from sklearn.metrics import (
    accuracy_score, mean_absolute_error,
    silhouette_score)
from sklearn.model_selection import train_test_split
from mlprodict.onnxrt import OnnxInference
from mlprodict.onnx_conv import (
    to_onnx, register_rewritten_operators, register_converters)
from mlprodict.onnxrt.validate.validate_benchmark import make_n_rows
from mlprodict.onnxrt.validate.validate_problems import _modify_dimension
from mlprodict.onnxrt.optim import onnx_statistics
from mlprodict.tools.asv_options_helper import (
    expand_onnx_options, get_opset_number_from_onnx,
    get_ir_version_from_onnx, version2number)
from mlprodict.tools.model_info import set_random_state


class _CommonAsvSklBenchmark:
    """
    Common tests to all benchmarks testing converted
    :epkg:`scikit-learn` models. See `benchmark attributes
    <https://asv.readthedocs.io/en/stable/benchmarks.html#general>`_.
    """

    # Part which changes.
    # params and param_names may be changed too.

    params = [
        ['skl', 'pyrtc', 'ort'],  # values for runtime
        [1, 10, 100, 10000],  # values for N
        [4, 20],  # values for nf
        [get_opset_number_from_onnx()],  # values for opset
        ["float", "double"],  # values for dtype
        [None],  # values for optim
    ]
    param_names = ['rt', 'N', 'nf', 'opset', 'dtype', 'optim']
    chk_method_name = None
    version = datetime.now().isoformat()
    pretty_source = "disabled"

    par_ydtype = numpy.int64
    par_dofit = True
    par_convopts = None

    def _create_model(self):  # pragma: no cover
        raise NotImplementedError("This method must be overwritten.")

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):  # pragma: no cover
        raise NotImplementedError("This method must be overwritten.")

    def _score_metric(self, X, y_exp, y_pred):  # pragma: no cover
        raise NotImplementedError("This method must be overwritten.")

    def _optimize_onnx(self, onx):
        return onx

    def _get_xdtype(self, dtype):
        if dtype in ('float', numpy.float32):
            return numpy.float32
        elif dtype in ('double', '64', 64, numpy.float64):
            return numpy.float64
        raise ValueError(  # pragma: no cover
            "Unknown dtype '{}'.".format(dtype))

    def _get_dataset(self, nf, dtype):
        xdtype = self._get_xdtype(dtype)
        data = load_iris()
        X, y = data.data, data.target
        state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
        rnd = state.randn(*X.shape) / 3
        X += rnd
        X = _modify_dimension(X, nf)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, random_state=42)
        Xt = X_test.astype(xdtype)
        yt = y_test.astype(self.par_ydtype)
        return (X_train, y_train), (Xt, yt)

    def _to_onnx(self, model, X, opset, dtype, optim):
        if optim is None or len(optim) == 0:
            options = self.par_convopts
        elif self.par_convopts and len(self.par_convopts) > 0:
            raise NotImplementedError(  # pragma: no cover
                "Conflict between par_convopts={} and optim={}".format(
                    self.par_convopts, optim))
        else:
            # Expand common onnx options, see _nick_name_options.
            options = expand_onnx_options(model, optim)

        return to_onnx(model, X, options=options, target_opset=opset)

    def _create_onnx_inference(self, onx, runtime):
        if 'onnxruntime' in runtime:
            old = onx.ir_version
            onx.ir_version = get_ir_version_from_onnx()
        else:
            old = None

        try:
            res = OnnxInference(onx, runtime=runtime)
        except RuntimeError as e:  # pragma: no cover
            if "[ONNXRuntimeError]" in str(e):
                return RuntimeError("onnxruntime fails due to {}".format(str(e)))
            raise e
        if old is not None:
            onx.ir_version = old
        return res

    # Part which does not change.

    def _check_rt(self, rt, meth):
        """
        Checks that runtime has the appropriate method.
        """
        if rt is None:
            raise ValueError("rt cannot be empty.")  # pragma: no cover
        if not hasattr(rt, meth):
            raise TypeError(  # pragma: no cover
                "rt of type %r has no method %r." % (type(rt), meth))

    def runtime_name(self, runtime):
        """
        Returns the runtime shortname.
        """
        if runtime == 'skl':
            name = runtime
        elif runtime == 'ort':
            name = 'onnxruntime1'
        elif runtime == 'ort2':
            name = 'onnxruntime2'  # pragma: no cover
        elif runtime == 'pyrt':
            name = 'python'
        elif runtime == 'pyrtc':
            name = 'python_compiled'
        else:
            raise ValueError(  # pragma: no cover
                "Unknown runtime '{}'.".format(runtime))
        return name

    def _name(self, nf, opset, dtype):
        last = 'cache-{}-nf{}-op{}-dt{}.pickle'.format(
            self.__class__.__name__, nf, opset, dtype)
        return last

    def setup_cache(self):
        "asv API"
        for dtype in self.params[4]:
            for opv in self.params[3]:
                for nf in self.params[2]:
                    (X_train, y_train), (X, y) = self._get_dataset(nf, dtype)
                    model = self._create_model()
                    if self.par_dofit:
                        set_random_state(model)
                        model.fit(X_train, y_train)
                    stored = {'model': model, 'X': X, 'y': y}
                    filename = self._name(nf, opv, dtype)
                    with open(filename, "wb") as f:
                        pickle.dump(stored, f)
                    if not os.path.exists(filename):
                        raise RuntimeError(  # pragma: no cover
                            "Unable to dump model %r into %r." % (
                                model, filename))

    def setup(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        logger = getLogger('skl2onnx')
        logger.disabled = True
        register_converters()
        register_rewritten_operators()
        with open(self._name(nf, opset, dtype), "rb") as f:
            stored = pickle.load(f)
        self.stored = stored
        self.model = stored['model']
        self.X, self.y = make_n_rows(stored['X'], N, stored['y'])
        onx, rt_, rt_fct_, rt_fct_track_ = self._create_onnx_and_runtime(
            runtime, self.model, self.X, opset, dtype, optim)
        self.onx = onx
        setattr(self, "rt_" + runtime, rt_)
        setattr(self, "rt_fct_" + runtime, rt_fct_)
        setattr(self, "rt_fct_track_" + runtime, rt_fct_track_)
        set_config(assume_finite=True)

    def time_predict(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        return getattr(self, "rt_fct_" + runtime)(self.X)

    def peakmem_predict(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        return getattr(self, "rt_fct_" + runtime)(self.X)

    def track_score(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        yp = getattr(self, "rt_fct_track_" + runtime)(self.X)
        return self._score_metric(self.X, self.y, yp)

    def track_onnxsize(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        return len(self.onx.SerializeToString())

    def track_nbnodes(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        stats = onnx_statistics(self.onx)
        return stats.get('nnodes', 0)

    def track_vmlprodict(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        from mlprodict import __version__
        return version2number(__version__)

    def track_vsklearn(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        from sklearn import __version__
        return version2number(__version__)

    def track_vort(self, runtime, N, nf, opset, dtype, optim):
        "asv API"
        try:
            from onnxruntime import __version__
            return version2number(__version__)
        except ImportError:  # pragma: no cover
            return 0

    def check_method_name(self, method_name):
        "Does some verifications. Fails if inconsistencies."
        if getattr(self, 'chk_method_name', None) not in (None, method_name):
            raise RuntimeError(  # pragma: no cover
                "Method name must be '{}'.".format(method_name))
        if getattr(self, 'chk_method_name', None) is None:
            raise RuntimeError(  # pragma: no cover
                "Unable to check that the method name is correct "
                "(expected is '{}')".format(
                    method_name))


class _CommonAsvSklBenchmarkClassifier(_CommonAsvSklBenchmark):
    """
    Common class for a classifier.
    """
    chk_method_name = 'predict_proba'

    def _score_metric(self, X, y_exp, y_pred):
        return accuracy_score(y_exp, y_pred)

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('predict_proba')
        onx_ = self._to_onnx(model, X, opset, dtype, optim)
        onx = self._optimize_onnx(onx_)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.predict_proba(X)
            rt_fct_track_ = lambda X: model.predict(X)
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda pX: rt_.run({'X': pX})
            rt_fct_track_ = lambda pX: rt_fct_(pX)['output_label']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkClassifierRawScore(_CommonAsvSklBenchmark):
    """
    Common class for a classifier.
    """
    chk_method_name = 'decision_function'

    def _score_metric(self, X, y_exp, y_pred):
        return accuracy_score(y_exp, y_pred)

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('decision_function')
        onx_ = self._to_onnx(model, X, opset, dtype, optim)
        onx = self._optimize_onnx(onx_)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.decision_function(X)
            rt_fct_track_ = lambda X: model.predict(X)
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda X: rt_.run({'X': X})
            rt_fct_track_ = lambda X: rt_fct_(X)['output_label']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkClustering(_CommonAsvSklBenchmark):
    """
    Common class for a clustering algorithm.
    """
    chk_method_name = 'predict'

    def _score_metric(self, X, y_exp, y_pred):
        if X.shape[0] == 1:
            return 0.  # pragma: no cover
        elif set(y_pred) == 1:
            return 0.  # pragma: no cover
        return silhouette_score(X, y_pred)

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('predict')
        onx_ = self._to_onnx(model, X, opset, dtype, optim)
        onx = self._optimize_onnx(onx_)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.predict(X.astype(numpy.float64))
            rt_fct_track_ = lambda X: model.predict(X.astype(numpy.float64))
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda X: rt_.run({'X': X})
            rt_fct_track_ = lambda X: rt_fct_(X)['label']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkMultiClassifier(_CommonAsvSklBenchmark):
    """
    Common class for a multi-classifier.
    """
    chk_method_name = 'predict_proba'

    def _get_dataset(self, nf, dtype):
        xdtype = self._get_xdtype(dtype)
        data = load_iris()
        X, y = data.data, data.target
        state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
        rnd = state.randn(*X.shape) / 3
        X += rnd
        nbclass = len(set(y))
        y_ = numpy.zeros((y.shape[0], nbclass), dtype=y.dtype)
        for i, vy in enumerate(y):
            y_[i, vy] = 1
        y = y_
        X = _modify_dimension(X, nf)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, random_state=42)
        X = X_test.astype(xdtype)
        y = y_test.astype(self.par_ydtype)
        return (X_train, y_train), (X, y)

    def _score_metric(self, X, y_exp, y_pred):
        return accuracy_score(y_exp.ravel(), y_pred.ravel())

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('predict_proba')
        onx_ = self._to_onnx(model, X, opset, dtype, optim)
        onx = self._optimize_onnx(onx_)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.predict_proba(X)
            rt_fct_track_ = lambda X: model.predict(X)
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda X: rt_.run({'X': X})
            rt_fct_track_ = lambda X: rt_fct_(X)['output_label']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkOutlier(_CommonAsvSklBenchmark):
    """
    Common class for outlier detection.
    """
    chk_method_name = 'predict'

    def _score_metric(self, X, y_exp, y_pred):
        return numpy.sum(y_pred) / y_pred.shape[0]

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('predict')
        onx_ = self._to_onnx(model, X, opset, dtype, optim)
        onx = self._optimize_onnx(onx_)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.predict(X)
            rt_fct_track_ = lambda X: model.predict(X)
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda X: rt_.run({'X': X})
            rt_fct_track_ = lambda X: rt_fct_(X)['scores']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkRegressor(_CommonAsvSklBenchmark):
    """
    Common class for a regressor.
    """
    chk_method_name = 'predict'

    def _score_metric(self, X, y_exp, y_pred):
        return mean_absolute_error(y_exp, y_pred)

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('predict')
        onx = self._to_onnx(model, X, opset, dtype, optim)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.predict(X)
            rt_fct_track_ = lambda X: model.predict(X)
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda X: rt_.run({'X': X})
            rt_fct_track_ = lambda X: rt_fct_(X)['variable']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkTrainableTransform(_CommonAsvSklBenchmark):
    """
    Common class for a trainable transformer.
    """
    chk_method_name = 'transform'

    def _score_metric(self, X, y_exp, y_pred):
        return numpy.sum(y_pred) / y_pred.shape[0]

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('transform')
        onx_ = self._to_onnx(model, X, opset, dtype, optim)
        onx = self._optimize_onnx(onx_)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.transform(X)
            rt_fct_track_ = lambda X: model.transform(X)
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda X: rt_.run({'X': X})
            rt_fct_track_ = lambda X: rt_fct_(X)['variable']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkTransform(_CommonAsvSklBenchmark):
    """
    Common class for a transformer.
    """
    chk_method_name = 'transform'

    def _score_metric(self, X, y_exp, y_pred):
        return numpy.sum(y_pred) / y_pred.shape[0]

    def _create_onnx_and_runtime(self, runtime, model, X, opset, dtype, optim):
        self.check_method_name('transform')
        onx_ = self._to_onnx(model, X, opset, dtype, optim)
        onx = self._optimize_onnx(onx_)
        name = self.runtime_name(runtime)
        if name == 'skl':
            rt_ = None
            rt_fct_ = lambda X: model.transform(X)
            rt_fct_track_ = lambda X: model.transform(X)
        else:
            rt_ = self._create_onnx_inference(onx, name)
            self._check_rt(rt_, 'run')
            rt_fct_ = lambda X: rt_.run({'X': X})
            rt_fct_track_ = lambda X: rt_fct_(X)['variable']
        return onx, rt_, rt_fct_, rt_fct_track_


class _CommonAsvSklBenchmarkTransformPositive(_CommonAsvSklBenchmarkTransform):
    """
    Common class for a transformer for positive features.
    """
    chk_method_name = 'transform'

    def _get_dataset(self, nf, dtype):
        xdtype = self._get_xdtype(dtype)
        data = load_iris()
        X, y = data.data, data.target
        state = numpy.random.RandomState(seed=34)  # pylint: disable=E1101
        rnd = state.randn(*X.shape) / 3
        X += rnd
        X = _modify_dimension(X, nf)
        X = numpy.abs(X)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, random_state=42)
        X = X_test.astype(xdtype)
        y = y_test.astype(self.par_ydtype)
        return (X_train, y_train), (X, y)
